# -*- coding: utf-8 -*-

# Copyright © 2012-2024 Roberto Alsina and others.

# Permission is hereby granted, free of charge, to any
# person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the
# Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice
# shall be included in all copies or substantial portions of
# the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
# OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""The Post class."""

import io
import datetime
import hashlib
import json
import os
import re
from collections import defaultdict
from math import ceil  # for reading time feature
from urllib.parse import urljoin

import dateutil.tz
import lxml.html
import natsort
from blinker import signal

# for tearDown with _reload we cannot use 'from import' to get forLocaleBorg
import nikola.utils
from . import metadata_extractors
from . import utils
from .utils import (
    current_time,
    Functionary,
    LOGGER,
    LocaleBorg,
    slugify,
    to_datetime,
    demote_headers,
    get_translation_candidate,
    map_metadata,
    bool_from_meta,
)

try:
    import pyphen
except ImportError:
    pyphen = None


__all__ = ('Post',)

TEASER_REGEXP = re.compile(r'<!--\s*(TEASER_END|END_TEASER)(:(.+))?\s*-->', re.IGNORECASE)


class Post(object):
    """Represent a blog post or site page."""

    _prev_post = None
    _next_post = None
    is_draft = False
    is_private = False
    _is_two_file = None
    _reading_time = None
    _remaining_reading_time = None
    _paragraph_count = None
    _remaining_paragraph_count = None
    post_status = 'published'
    has_oldstyle_metadata_tags = False

    def __init__(
        self,
        source_path,
        config,
        destination,
        use_in_feeds,
        messages,
        template_name,
        compiler,
        destination_base=None,
        metadata_extractors_by=None
    ):
        """Initialize post.

        The source path is the user created post file. From it we calculate
        the meta file, as well as any translations available, and
        the .html fragment file path.

        destination_base must be None or a TranslatableSetting instance. If
        specified, it will be prepended to the destination path.
        """
        self._load_config(config)
        self._set_paths(source_path)

        self.compiler = compiler
        self.is_post = use_in_feeds
        self.messages = messages
        self._template_name = template_name
        self.compile_html = self.compiler.compile
        self.demote_headers = self.compiler.demote_headers and self.config['DEMOTE_HEADERS']
        self._dependency_file_fragment = defaultdict(list)
        self._dependency_file_page = defaultdict(list)
        self._dependency_uptodate_fragment = defaultdict(list)
        self._dependency_uptodate_page = defaultdict(list)
        self._depfile = defaultdict(list)
        if metadata_extractors_by is None:
            self.metadata_extractors_by = {'priority': {}, 'source': {}}
        else:
            self.metadata_extractors_by = metadata_extractors_by

        self._set_translated_to()
        self._set_folders(destination, destination_base)

        # Load default metadata
        default_metadata, default_used_extractor = get_meta(self, lang=None)
        self.meta = Functionary(lambda: None, self.default_lang)
        self.used_extractor = Functionary(lambda: None, self.default_lang)
        self.meta[self.default_lang] = default_metadata
        self.used_extractor[self.default_lang] = default_used_extractor

        self._set_date(default_metadata)

        # These are the required metadata fields
        if 'title' not in default_metadata or 'slug' not in default_metadata:
            raise ValueError("You must set a title (found '{0}') and a slug (found '{1}')! "
                             "[in file {2}]".format(default_metadata.get('title', None),
                                                    default_metadata.get('slug', None),
                                                    source_path))

        if 'type' not in default_metadata:
            default_metadata['type'] = 'text'

        self._load_translated_metadata(default_metadata)
        self._load_data()
        self.__migrate_section_to_category()
        self._set_tags()

        self.publish_later = False if self.current_time is None else self.date >= self.current_time

        # While draft comes from the tags, it's not really a tag
        self.use_in_feeds = self.is_post and not self.is_draft and not self.is_private and not self.publish_later

        # Allow overriding URL_TYPE via meta
        # The check is done here so meta dicts won’t change inside of
        # generic_post_renderer
        self.url_type = self.meta('url_type') or None
        # Register potential extra dependencies
        self.compiler.register_extra_dependencies(self)

    def _load_config(self, config):
        """Set members to configured values."""
        self.config = config
        if self.config['FUTURE_IS_NOW']:
            self.current_time = None
        else:
            self.current_time = current_time(self.config['__tzinfo__'])
        self.base_url = self.config['BASE_URL']
        self.strip_indexes = self.config['STRIP_INDEXES']
        self.index_file = self.config['INDEX_FILE']
        self.pretty_urls = self.config['PRETTY_URLS']
        self.default_lang = self.config['DEFAULT_LANG']
        self.translations = self.config['TRANSLATIONS']
        self.skip_untranslated = not self.config['SHOW_UNTRANSLATED_POSTS']
        self._default_preview_image = self.config['DEFAULT_PREVIEW_IMAGE']
        self.types_to_hide_title = self.config['TYPES_TO_HIDE_TITLE']

    def _set_tags(self):
        """Set post tags."""
        self._tags = {}
        for lang in self.translated_to:
            if isinstance(self.meta[lang]['tags'], (list, tuple, set)):
                _tag_list = self.meta[lang]['tags']
            else:
                _tag_list = self.meta[lang]['tags'].split(',')
            self._tags[lang] = natsort.natsorted(
                list(set([x.strip() for x in _tag_list])),
                alg=natsort.ns.F | natsort.ns.IC)
            self._tags[lang] = [t for t in self._tags[lang] if t]

            status = self.meta[lang].get('status')
            if status:
                if status == 'published':
                    pass  # already set before, mixing published + something else should result in the other thing
                elif status == 'featured':
                    self.post_status = status
                elif status == 'private':
                    self.post_status = status
                    self.is_private = True
                elif status == 'draft':
                    self.post_status = status
                    self.is_draft = True
                else:
                    LOGGER.warning(('The post "{0}" has the unknown status "{1}". '
                                    'Valid values are "published", "featured", "private" and "draft".').format(self.source_path, status))

            if self.config['WARN_ABOUT_TAG_METADATA']:
                show_warning = False
                if 'draft' in [_.lower() for _ in self._tags[lang]]:
                    LOGGER.warning('The post "{0}" uses the "draft" tag.'.format(self.source_path))
                    show_warning = True
                if 'private' in self._tags[lang]:
                    LOGGER.warning('The post "{0}" uses the "private" tag.'.format(self.source_path))
                    show_warning = True
                if 'mathjax' in self._tags[lang]:
                    LOGGER.warning('The post "{0}" uses the "mathjax" tag.'.format(self.source_path))
                    show_warning = True
                if show_warning:
                    LOGGER.warning('It is suggested that you convert special tags to metadata and set '
                                   'USE_TAG_METADATA to False. You can use the upgrade_metadata_v8 '
                                   'command plugin for conversion (install with: nikola plugin -i '
                                   'upgrade_metadata_v8). Change the WARN_ABOUT_TAG_METADATA '
                                   'configuration to disable this warning.')
            if self.config['USE_TAG_METADATA']:
                if 'draft' in [_.lower() for _ in self._tags[lang]]:
                    self.is_draft = True
                    LOGGER.debug('The post "{0}" is a draft.'.format(self.source_path))
                    self._tags[lang].remove('draft')
                    self.post_status = 'draft'
                    self.has_oldstyle_metadata_tags = True

                if 'private' in self._tags[lang]:
                    self.is_private = True
                    LOGGER.debug('The post "{0}" is private.'.format(self.source_path))
                    self._tags[lang].remove('private')
                    self.post_status = 'private'
                    self.has_oldstyle_metadata_tags = True

                if 'mathjax' in self._tags[lang]:
                    self.has_oldstyle_metadata_tags = True

    def _set_paths(self, source_path):
        """Set the various paths and the post_name.

        TODO: WTF is all this.
        """
        self.source_path = source_path  # posts/blah.txt
        self.post_name = os.path.splitext(source_path)[0]  # posts/blah
        _relpath = os.path.relpath(self.post_name)
        if _relpath != self.post_name:
            self.post_name = _relpath.replace('..' + os.sep, '__dotdot__' + os.sep)
        # cache[\/]posts[\/]blah.html
        self.base_path = os.path.join(self.config['CACHE_FOLDER'], self.post_name + ".html")
        # cache/posts/blah.html
        self._base_path = self.base_path.replace('\\', '/')
        self.metadata_path = self.post_name + ".meta"  # posts/blah.meta

    def _set_translated_to(self):
        """Find post's translations."""
        self.translated_to = set([])
        for lang in self.translations:
            if os.path.isfile(get_translation_candidate(self.config, self.source_path, lang)):
                self.translated_to.add(lang)

        # If we don't have anything in translated_to, the file does not exist
        if not self.translated_to and os.path.isfile(self.source_path):
            raise Exception(("Could not find translations for {}, check your "
                            "TRANSLATIONS_PATTERN").format(self.source_path))
        elif not self.translated_to:
            raise Exception(("Cannot use {} (not a file, perhaps a broken "
                            "symbolic link?)").format(self.source_path))

    def _set_folders(self, destination, destination_base):
        """Compose destination paths."""
        self.folder_relative = destination
        self.folder_base = destination_base

        if self.folder_base is not None:
            # Use translatable destination folders
            self.folders = {}
            for lang in self.config['TRANSLATIONS']:
                if os.path.isabs(self.folder_base(lang)):  # Issue 2982
                    self.folder_base[lang] = os.path.relpath(self.folder_base(lang), '/')
                self.folders[lang] = os.path.normpath(os.path.join(self.folder_base(lang), self.folder_relative))
        else:
            # Old behavior (non-translatable destination path, normalized by scanner)
            self.folders = {lang: self.folder_relative for lang in self.config['TRANSLATIONS'].keys()}
        self.folder = self.folders[self.default_lang]

    def __migrate_section_to_category(self):
        """TODO: remove in v9."""
        for lang, meta in self.meta.items():
            # Migrate section to category
            # TODO: remove in v9
            if 'section' in meta:
                if 'category' in meta:
                    LOGGER.warning("Post {0} has both 'category' and 'section' metadata. Section will be ignored.".format(self.source_path))
                else:
                    meta['category'] = meta['section']
                    LOGGER.info("Post {0} uses 'section' metadata, setting its value to 'category'".format(self.source_path))

            # Handle CATEGORY_DESTPATH_AS_DEFAULT
            if 'category' not in meta and self.config['CATEGORY_DESTPATH_AS_DEFAULT']:
                self.category_from_destpath = True
                if self.config['CATEGORY_DESTPATH_TRIM_PREFIX'] and self.folder_relative != '.':
                    category = self.folder_relative
                else:
                    category = self.folders[lang]
                category = category.replace(os.sep, '/')
                if self.config['CATEGORY_DESTPATH_FIRST_DIRECTORY_ONLY']:
                    category = category.split('/')[0]
                meta['category'] = self.config['CATEGORY_DESTPATH_NAMES'](lang).get(category, category)
            else:
                self.category_from_destpath = False

    def _load_data(self):
        """Load data field from metadata."""
        self.data = Functionary(lambda: None, self.default_lang)
        for lang in self.translations:
            if self.meta[lang].get('data') is not None:
                self.data[lang] = utils.load_data(self.meta[lang]['data'])
                self.register_depfile(self.meta[lang]['data'], lang=lang)

    def _load_translated_metadata(self, default_metadata):
        """Load metadata from all translation sources."""
        for lang in self.translations:
            if lang != self.default_lang:
                meta = defaultdict(lambda: '')
                meta.update(default_metadata)
                _meta, _extractors = get_meta(self, lang)
                meta.update(_meta)
                self.meta[lang] = meta
                self.used_extractor[lang] = _extractors

        if not self.is_translation_available(self.default_lang):
            # Special case! (Issue #373)
            # Fill default_metadata with stuff from the other languages
            for lang in sorted(self.translated_to):
                default_metadata.update(self.meta[lang])

    def _set_date(self, default_metadata):
        """Set post date/updated based on metadata and configuration."""
        if 'date' not in default_metadata and not self.is_post:
            # For pages we don't *really* need a date
            if self.config['__invariant__']:
                default_metadata['date'] = datetime.datetime(2013, 12, 31, 23, 59, 59, tzinfo=self.config['__tzinfo__'])
            else:
                default_metadata['date'] = datetime.datetime.fromtimestamp(
                    os.stat(self.source_path).st_ctime, dateutil.tz.tzutc()).astimezone(self.config['__tzinfo__'])

        # If time zone is set, build localized datetime.
        try:
            self.date = to_datetime(self.meta[self.default_lang]['date'], self.config['__tzinfo__'])
        except ValueError:
            if not self.meta[self.default_lang]['date']:
                msg = 'Missing date in file {}'.format(self.source_path)
            else:
                msg = "Invalid date '{0}' in file {1}".format(self.meta[self.default_lang]['date'], self.source_path)
            LOGGER.error(msg)
            raise ValueError(msg)

        if 'updated' not in default_metadata:
            default_metadata['updated'] = default_metadata.get('date', None)

        self.updated = to_datetime(default_metadata['updated'], self.config['__tzinfo__'])

    @property
    def hyphenate(self):
        """Post is hyphenated."""
        return bool(self.config['HYPHENATE'] or self.meta('hyphenate'))

    @property
    def is_two_file(self):
        """Post has a separate .meta file."""
        if self._is_two_file is None:
            return True
        return self._is_two_file

    @is_two_file.setter
    def is_two_file(self, value):
        """Set the is_two_file property, use with care.

        Caution: this MAY REWRITE THE POST FILE.
        Only should happen if you effectively *change* the value.

        Arguments:
            value {bool} -- Whether the post has a separate .meta file
        """
        # for lang in self.translated_to:

        if self._is_two_file is None:
            # Initial setting, this happens on post creation
            self._is_two_file = value
        elif value != self._is_two_file:
            # Changing the value, this means you are transforming a 2-file
            # into a 1-file or viceversa.
            if value and not self.compiler.supports_metadata:
                raise ValueError("Can't save metadata as 1-file using this compiler {}".format(self.compiler))
            for lang in self.translated_to:
                source = self.source(lang)
                meta = self.meta(lang)
                self._is_two_file = value
                self.save(lang=lang, source=source, meta=meta)
                if not value:  # Need to delete old meta file
                    meta_path = get_translation_candidate(self.config, self.metadata_path, lang)
                    if os.path.isfile(meta_path):
                        os.unlink(meta_path)

    def __repr__(self):
        """Provide a representation of the post object."""
        # Calculate a hash that represents most data about the post
        m = hashlib.md5()
        # source_path modification date (to avoid reading it)
        m.update(str(os.stat(self.source_path).st_mtime).encode('utf-8'))
        clean_meta = {}
        for k, v in self.meta.items():
            sub_meta = {}
            clean_meta[k] = sub_meta
            for kk, vv in v.items():
                if vv:
                    sub_meta[kk] = vv
        m.update(str(json.dumps(clean_meta, cls=utils.CustomEncoder, sort_keys=True)).encode('utf-8'))
        return '<Post: {0!r} {1}>'.format(self.source_path, m.hexdigest())

    def has_pretty_url(self, lang):
        """Check if this page has a pretty URL."""
        meta_value = bool_from_meta(self.meta[lang], 'pretty_url')

        if meta_value is None:
            # use PRETTY_URLS, unless the slug is 'index'
            return self.pretty_urls and self.meta[lang]['slug'] != 'index'
        else:
            # override with meta value
            return meta_value

    def _has_pretty_url(self, lang):
        """Check if this page has a pretty URL."""
        return self.has_pretty_url(lang)

    @property
    def has_math(self):
        """Return True if this post has has_math set to True or is a python notebook.

        Alternatively, it will return True if it has set the mathjax tag in the
        current language and the USE_TAG_METADATA config setting is True.
        """
        if self.compiler.name == 'ipynb':
            return True
        lang = nikola.utils.LocaleBorg().current_lang
        if self.is_translation_available(lang):
            if bool_from_meta(self.meta[lang], 'has_math'):
                return True
            if self.config['USE_TAG_METADATA']:
                return 'mathjax' in self.tags_for_language(lang)
        # If it has math in ANY other language, enable it. Better inefficient than broken.
        for lang in self.translated_to:
            if bool_from_meta(self.meta[lang], 'has_math'):
                return True
        if self.config['USE_TAG_METADATA']:
            return 'mathjax' in self.alltags
        return False

    @property
    def alltags(self):
        """Return ALL the tags for this post."""
        tags = []
        for l in self._tags:
            tags.extend(self._tags[l])
        return list(set(tags))

    def tags_for_language(self, lang):
        """Return tags for a given language."""
        if lang in self._tags:
            return self._tags[lang]
        elif lang not in self.translated_to and self.skip_untranslated:
            return []
        elif self.default_lang in self._tags:
            return self._tags[self.default_lang]
        else:
            return []

    @property
    def tags(self):
        """Return tags for the current language."""
        lang = nikola.utils.LocaleBorg().current_lang
        return self.tags_for_language(lang)

    @property
    def prev_post(self):
        """Return previous post."""
        lang = nikola.utils.LocaleBorg().current_lang
        rv = self._prev_post
        while self.skip_untranslated:
            if rv is None:
                break
            if rv.is_translation_available(lang):
                break
            rv = rv._prev_post
        return rv

    @prev_post.setter
    def prev_post(self, v):
        """Set previous post."""
        self._prev_post = v

    @property
    def next_post(self):
        """Return next post."""
        lang = nikola.utils.LocaleBorg().current_lang
        rv = self._next_post
        while self.skip_untranslated:
            if rv is None:
                break
            if rv.is_translation_available(lang):
                break
            rv = rv._next_post
        return rv

    @next_post.setter
    def next_post(self, v):
        """Set next post."""
        self._next_post = v

    @property
    def template_name(self):
        """Return template name for this post."""
        lang = nikola.utils.LocaleBorg().current_lang
        return self.meta[lang]['template'] or self._template_name

    def formatted_date(self, date_format, date=None):
        """Return the formatted date as string."""
        return utils.LocaleBorg().formatted_date(date_format, date if date else self.date)

    def formatted_updated(self, date_format):
        """Return the updated date as string."""
        return self.formatted_date(date_format, self.updated)

    def title(self, lang=None):
        """Return localized title.

        If lang is not specified, it defaults to the current language from
        templates, as set in LocaleBorg.
        """
        if lang is None:
            lang = nikola.utils.LocaleBorg().current_lang
        return self.meta[lang]['title']

    def author(self, lang=None) -> str:
        """Return localized author or BLOG_AUTHOR if unspecified.

        If lang is not specified, it defaults to the current language from
        templates, as set in LocaleBorg.
        """
        if lang is None:
            lang = nikola.utils.LocaleBorg().current_lang
        if self.meta[lang]['author']:
            author = self.meta[lang]['author']
        else:
            author = self.config['BLOG_AUTHOR'](lang)

        return author

    def authors(self, lang=None) -> list:
        """Return localized authors or BLOG_AUTHOR if unspecified.

        If lang is not specified, it defaults to the current language from
        templates, as set in LocaleBorg.
        """
        if lang is None:
            lang = nikola.utils.LocaleBorg().current_lang
        if self.meta[lang]['author']:
            author = [i.strip() for i in self.meta[lang]['author'].split(",")]
        else:
            author = [self.config['BLOG_AUTHOR'](lang)]

        return author

    def description(self, lang=None):
        """Return localized description."""
        if lang is None:
            lang = nikola.utils.LocaleBorg().current_lang
        return self.meta[lang]['description']

    def guid(self, lang=None):
        """Return localized GUID."""
        if lang is None:
            lang = nikola.utils.LocaleBorg().current_lang
        if self.meta[lang]['guid']:
            guid = self.meta[lang]['guid']
        else:
            guid = self.permalink(lang, absolute=True)

        return guid

    def add_dependency(self, dependency, add='both', lang=None):
        """Add a file dependency for tasks using that post.

        The ``dependency`` should be a string specifying a path, or a callable
        which returns such a string or a list of strings.

        The ``add`` parameter can be 'both', 'fragment' or 'page', to indicate
        that this dependency shall be used
         * when rendering the fragment to HTML ('fragment' and 'both'), or
         * when creating a page with parts of the ``Post`` embedded, which
           includes the HTML resulting from compiling the fragment ('page' or
           'both').

        If ``lang`` is not specified, this dependency is added for all languages.
        """
        if add not in {'fragment', 'page', 'both'}:
            raise Exception("Add parameter is '{0}', but must be either 'fragment', 'page', or 'both'.".format(add))
        if add == 'fragment' or add == 'both':
            self._dependency_file_fragment[lang].append((not isinstance(dependency, str), dependency))
        if add == 'page' or add == 'both':
            self._dependency_file_page[lang].append((not isinstance(dependency, str), dependency))

    def add_dependency_uptodate(self, dependency, is_callable=False, add='both', lang=None):
        """Add a dependency for task's ``uptodate`` for tasks using that post.

        This can be for example an ``utils.config_changed`` object, or a list of
        such objects.

        The ``is_callable`` parameter specifies whether ``dependency`` is a
        callable which generates an entry or a list of entries for the ``uptodate``
        list, or whether it is an entry which can directly be added (as a single
        object or a list of objects).

        The ``add`` parameter can be 'both', 'fragment' or 'page', to indicate
        that this dependency shall be used
         * when rendering the fragment to HTML ('fragment' and 'both'), or
         * when creating a page with parts of the ``Post`` embedded, which
           includes the HTML resulting from compiling the fragment ('page' or
           'both').

        If ``lang`` is not specified, this dependency is added for all languages.

        Example:

        post.add_dependency_uptodate(
            utils.config_changed({1: some_data}, 'uniqueid'), False, 'page')
        """
        if add == 'fragment' or add == 'both':
            self._dependency_uptodate_fragment[lang].append((is_callable, dependency))
        if add == 'page' or add == 'both':
            self._dependency_uptodate_page[lang].append((is_callable, dependency))

    def register_depfile(self, dep, dest=None, lang=None):
        """Register a dependency in the dependency file."""
        if not dest:
            dest = self.translated_base_path(lang)
        self._depfile[dest].append(dep)

    @staticmethod
    def write_depfile(dest, deps_list, post=None, lang=None):
        """Write a depfile for a given language."""
        if post is None or lang is None:
            deps_path = dest + '.dep'
        else:
            deps_path = post.compiler.get_dep_filename(post, lang)
        if deps_list or (post.compiler.use_dep_file if post else False):
            deps_list = [p for p in deps_list if p != dest]  # Don't depend on yourself (#1671)
            with io.open(deps_path, "w+", encoding="utf-8") as deps_file:
                deps_file.write('\n'.join(deps_list))
        else:
            if os.path.isfile(deps_path):
                os.unlink(deps_path)

    def _get_dependencies(self, deps_list):
        deps = []
        for dep in deps_list:
            if dep[0]:
                # callable
                result = dep[1]()
            else:
                # can add directly
                result = dep[1]
            # if result is a list, add its contents
            if isinstance(result, list):
                deps.extend(result)
            else:
                deps.append(result)
        return deps

    def deps(self, lang):
        """Return a list of file dependencies to build this post's page."""
        deps = []
        deps.append(self.base_path)
        deps.append(self.source_path)
        if os.path.exists(self.metadata_path):
            deps.append(self.metadata_path)
        if lang != self.default_lang:
            cand_1 = get_translation_candidate(self.config, self.source_path, lang)
            cand_2 = get_translation_candidate(self.config, self.base_path, lang)
            if os.path.exists(cand_1):
                deps.extend([cand_1, cand_2])
            cand_3 = get_translation_candidate(self.config, self.metadata_path, lang)
            if os.path.exists(cand_3):
                deps.append(cand_3)
        if self.meta('data', lang):
            deps.append(self.meta('data', lang))
        deps += self._get_dependencies(self._dependency_file_page[lang])
        deps += self._get_dependencies(self._dependency_file_page[None])
        return sorted(set(deps))

    def deps_uptodate(self, lang):
        """Return a list of uptodate dependencies to build this post's page.

        These dependencies should be included in ``uptodate`` for the task
        which generates the page.
        """
        deps = []
        deps += self._get_dependencies(self._dependency_uptodate_page[lang])
        deps += self._get_dependencies(self._dependency_uptodate_page[None])
        deps.append(utils.config_changed({1: sorted(self.compiler.config_dependencies)}, 'nikola.post.Post.deps_uptodate:compiler:' + self.source_path))
        return deps

    def compile(self, lang):
        """Generate the cache/ file with the compiled post."""
        dest = self.translated_base_path(lang)
        if not self.is_translation_available(lang) and not self.config['SHOW_UNTRANSLATED_POSTS']:
            return
        # Set the language to the right thing
        LocaleBorg().set_locale(lang)
        self.compile_html(
            self.translated_source_path(lang),
            dest,
            self.is_two_file,
            self,
            lang)
        Post.write_depfile(dest, self._depfile[dest], post=self, lang=lang)

        signal('compiled').send({
            'source': self.translated_source_path(lang),
            'dest': dest,
            'post': self,
            'lang': lang,
        })

        if self.publish_later:
            LOGGER.info('{0} is scheduled to be published in the future ({1})'.format(
                self.source_path, self.date))

    def fragment_deps(self, lang):
        """Return a list of dependencies to build this post's fragment."""
        deps = [self.source_path]
        if os.path.isfile(self.metadata_path):
            deps.append(self.metadata_path)
        lang_deps = []
        if lang != self.default_lang:
            lang_deps = [get_translation_candidate(self.config, d, lang) for d in deps]
            deps += lang_deps
        deps = [d for d in deps if os.path.exists(d)]
        deps += self._get_dependencies(self._dependency_file_fragment[lang])
        deps += self._get_dependencies(self._dependency_file_fragment[None])
        return sorted(deps)

    def fragment_deps_uptodate(self, lang):
        """Return a list of file dependencies to build this post's fragment."""
        deps = []
        deps += self._get_dependencies(self._dependency_uptodate_fragment[lang])
        deps += self._get_dependencies(self._dependency_uptodate_fragment[None])
        deps.append(utils.config_changed({1: sorted(self.compiler.config_dependencies)}, 'nikola.post.Post.deps_uptodate:compiler:' + self.source_path))
        return deps

    def is_translation_available(self, lang):
        """Return True if the translation actually exists."""
        return lang in self.translated_to

    def translated_source_path(self, lang):
        """Return path to the translation's source file."""
        if lang in self.translated_to:
            if lang == self.default_lang:
                return self.source_path
            else:
                return get_translation_candidate(self.config, self.source_path, lang)
        elif lang != self.default_lang:
            return self.source_path
        else:
            return get_translation_candidate(self.config, self.source_path, sorted(self.translated_to)[0])

    def translated_base_path(self, lang):
        """Return path to the translation's base_path file."""
        return get_translation_candidate(self.config, self.base_path, lang)

    def _translated_file_path(self, lang):
        """Get path to a post's translation.

        Returns path to the translation's file, or to as good a file as it can
        plus "real" language of the text.
        """
        if lang in self.translated_to:
            if lang == self.default_lang:
                return self.base_path, lang
            else:
                return get_translation_candidate(self.config, self.base_path, lang), lang
        elif lang != self.default_lang:
            return self.base_path, self.default_lang
        else:
            real_lang = sorted(self.translated_to)[0]
            return get_translation_candidate(self.config, self.base_path, real_lang), real_lang

    def write_metadata(self, lang=None):
        """Save the post's metadata.

        Keep in mind that this will save either in the
        post file or in a .meta file, depending on self.is_two_file.

        metadata obtained from filenames or document contents will
        be superseded by this, and becomes inaccessible.

        Post contents will **not** be modified.

        If you write to a language not in self.translated_to
        an exception will be raised.

        Remember to scan_posts(really=True) after you update metadata if
        you want the rest of the system to know about the change.
        """
        if lang is None:
            lang = nikola.utils.LocaleBorg().current_lang
        if lang not in self.translated_to:
            raise ValueError("Can't save post metadata to language [{}] it's not translated to.".format(lang))

        source = self.source(lang)
        source_path = self.translated_source_path(lang)
        metadata = self.meta[lang]
        self.compiler.create_post(source_path, content=source, onefile=not self.is_two_file, is_page=not self.is_post, **metadata)

    def save(self, lang=None, source=None, meta=None):
        """Write post source to disk.

        Use this with utmost care, it may wipe out a post.

        Keyword Arguments:
            lang str -- Language for this source. If set to None,
                use current language.
            source str -- The source text for the post in the
                language. If set to None, use current source for
                this language.
            meta dict -- Metadata for this language, if not set,
                use current metadata for this language.
        """
        if lang is None:
            lang = nikola.utils.LocaleBorg().current_lang
        if source is None:
            source = self.source(lang)
        if meta is None:
            metadata = self.meta[lang]
        source_path = self.translated_source_path(lang)
        metadata = self.meta[lang]
        self.compiler.create_post(source_path, content=source, onefile=not self.is_two_file, is_page=not self.is_post, **metadata)

    def source(self, lang=None):
        """Read the post and return its source."""
        if lang is None:
            lang = nikola.utils.LocaleBorg().current_lang

        source = self.translated_source_path(lang)
        with open(source, 'r', encoding='utf-8-sig') as inf:
            data = inf.read()
        if self.is_two_file:  # Metadata is not here
            source_data = data
        else:
            source_data = self.compiler.split_metadata(data, self, lang)[1]
        return source_data

    def text(self, lang=None, teaser_only=False, strip_html=False, show_read_more_link=True,
             feed_read_more_link=False, feed_links_append_query=None):
        """Read the post file for that language and return its compiled contents.

        teaser_only=True breaks at the teaser marker and returns only the teaser.
        strip_html=True removes HTML tags
        show_read_more_link=False does not add the Read more... link
        feed_read_more_link=True uses FEED_READ_MORE_LINK instead of INDEX_READ_MORE_LINK
        lang=None uses the last used to set locale

        All links in the returned HTML will be relative.
        The HTML returned is a bare fragment, not a full document.
        """
        if lang is None:
            lang = nikola.utils.LocaleBorg().current_lang
        file_name, real_lang = self._translated_file_path(lang)

        # Yes, we compile it and screw it.
        # This may be controversial, but the user (or someone) is asking for the post text
        # and the post should not just refuse to give it.
        if not os.path.isfile(file_name):
            self.compile(lang)

        with io.open(file_name, "r", encoding="utf-8-sig") as post_file:
            data = post_file.read().strip()

        if self.compiler.extension() == '.php':
            return data
        try:
            document = lxml.html.fragment_fromstring(data, "body")
        except lxml.etree.ParserError as e:
            # if we don't catch this, it breaks later (Issue #374)
            if str(e) == "Document is empty":
                return ""
            # let other errors raise
            raise
        base_url = self.permalink(lang=lang)
        document.make_links_absolute(base_url)

        if self.hyphenate:
            hyphenate(document, real_lang)

        data = utils.html_tostring_fragment(document)

        if teaser_only:
            teaser_regexp = self.config.get('TEASER_REGEXP', TEASER_REGEXP)
            teaser = teaser_regexp.split(data)[0]
            if teaser != data:
                if not strip_html and show_read_more_link:
                    if teaser_regexp.search(data).groups()[-1]:
                        teaser_text = teaser_regexp.search(data).groups()[-1]
                    else:
                        teaser_text = self.messages[lang]["Read more"]
                    l = self.config['FEED_READ_MORE_LINK'](lang) if feed_read_more_link else self.config['INDEX_READ_MORE_LINK'](lang)
                    teaser += l.format(
                        link=self.permalink(lang, query=feed_links_append_query),
                        read_more=teaser_text,
                        min_remaining_read=self.messages[lang]["%d min remaining to read"] % (self.remaining_reading_time),
                        reading_time=self.reading_time,
                        remaining_reading_time=self.remaining_reading_time,
                        paragraph_count=self.paragraph_count,
                        remaining_paragraph_count=self.remaining_paragraph_count,
                        post_title=self.title(lang))
                # This closes all open tags and sanitizes the broken HTML
                document = lxml.html.fromstring(teaser)
                data = utils.html_tostring_fragment(document)

        if data and strip_html:
            try:
                # Not all posts have a body. For example, you may have a page statically defined in the template that does not take content as input.
                content = lxml.html.fromstring(data)
                data = content.text_content().strip()  # No whitespace wanted.
            except (lxml.etree.ParserError, ValueError):
                data = ""
        elif data:
            if self.demote_headers:
                # see above
                try:
                    document = lxml.html.fragment_fromstring(data, "body")
                    demote_headers(document, self.demote_headers)
                    data = utils.html_tostring_fragment(document)
                except (lxml.etree.ParserError, IndexError):
                    pass

        return data

    @property
    def reading_time(self):
        """Return reading time based on length of text."""
        if self._reading_time is None:
            text = self.text(strip_html=True)
            words_per_minute = 220
            words = len(text.split())
            markup = lxml.html.fromstring(self.text(strip_html=False))
            embeddables = [".//img", ".//picture", ".//video", ".//audio", ".//object", ".//iframe"]
            media_time = 0
            for embedded in embeddables:
                media_time += (len(markup.findall(embedded)) * 0.33)  # +20 seconds
            self._reading_time = int(ceil((words / words_per_minute) + media_time)) or 1
        return self._reading_time

    @property
    def remaining_reading_time(self):
        """Remaining reading time based on length of text (does not include teaser)."""
        if self._remaining_reading_time is None:
            text = self.text(teaser_only=True, strip_html=True)
            words_per_minute = 220
            words = len(text.split())
            self._remaining_reading_time = self.reading_time - int(ceil(words / words_per_minute)) or 1
        return self._remaining_reading_time

    @property
    def paragraph_count(self):
        """Return the paragraph count for this post."""
        if self._paragraph_count is None:
            # duplicated with Post.text()
            lang = nikola.utils.LocaleBorg().current_lang
            file_name, _ = self._translated_file_path(lang)
            with io.open(file_name, "r", encoding="utf-8-sig") as post_file:
                data = post_file.read().strip()
            try:
                document = lxml.html.fragment_fromstring(data, "body")
            except lxml.etree.ParserError as e:
                # if we don't catch this, it breaks later (Issue #374)
                if str(e) == "Document is empty":
                    return ""
                # let other errors raise
                raise

            # output is a float, for no real reason at all
            self._paragraph_count = int(document.xpath('count(//p)'))
        return self._paragraph_count

    @property
    def remaining_paragraph_count(self):
        """Return the remaining paragraph count for this post (does not include teaser)."""
        if self._remaining_paragraph_count is None:
            try:
                # Just asking self.text() is easier here.
                document = lxml.html.fragment_fromstring(self.text(teaser_only=True, show_read_more_link=False), "body")
            except lxml.etree.ParserError as e:
                # if we don't catch this, it breaks later (Issue #374)
                if str(e) == "Document is empty":
                    return ""
                # let other errors raise
                raise

            self._remaining_paragraph_count = self.paragraph_count - int(document.xpath('count(//p)'))
        return self._remaining_paragraph_count

    def source_link(self, lang=None):
        """Return absolute link to the post's source."""
        ext = self.source_ext(True)
        link = "/" + self.destination_path(lang=lang, extension=ext, sep='/')
        link = utils.encodelink(link)
        return link

    def destination_path(self, lang=None, extension='.html', sep=os.sep):
        """Destination path for this post, relative to output/.

        If lang is not specified, it's the current language.
        Extension is used in the path if specified.
        """
        if lang is None:
            lang = nikola.utils.LocaleBorg().current_lang
        folder = self.folders[lang]
        if self.has_pretty_url(lang):
            path = os.path.join(self.translations[lang],
                                folder, self.meta[lang]['slug'], 'index' + extension)
        else:
            path = os.path.join(self.translations[lang],
                                folder, self.meta[lang]['slug'] + extension)
        if sep != os.sep:
            path = path.replace(os.sep, sep)
        if path.startswith('./'):
            path = path[2:]
        return path

    def permalink(self, lang=None, absolute=False, extension='.html', query=None):
        """Return permalink for a post."""
        if lang is None:
            lang = nikola.utils.LocaleBorg().current_lang

        # Let compilers override extension (e.g. the php compiler)
        if self.compiler.extension() != '.html':
            extension = self.compiler.extension()

        pieces = self.translations[lang].split(os.sep)
        pieces += self.folders[lang].split(os.sep)
        if self.has_pretty_url(lang):
            pieces += [self.meta[lang]['slug'], 'index' + extension]
        else:
            pieces += [self.meta[lang]['slug'] + extension]
        pieces = [_f for _f in pieces if _f and _f != '.']
        link = '/' + '/'.join(pieces)
        if absolute:
            link = urljoin(self.base_url, link[1:])
        index_len = len(self.index_file)
        if self.strip_indexes and link[-(1 + index_len):] == '/' + self.index_file:
            link = link[:-index_len]
        if query:
            link = link + "?" + query
        link = utils.encodelink(link)
        return link

    @property
    def previewimage(self, lang=None):
        """Return the previewimage path."""
        if lang is None:
            lang = nikola.utils.LocaleBorg().current_lang

        image_path = self.meta[lang]['previewimage']
        if not image_path:
            image_path = self._default_preview_image

        if not image_path or image_path.startswith("/"):
            # Paths starting with slashes are expected to be root-relative, pass them directly.
            return image_path
        # Other paths are relative to the permalink. The path will be made prettier by the URL replacer later.
        return urljoin(self.permalink(lang), image_path)

    def source_ext(self, prefix=False):
        """Return the source file extension.

        If `prefix` is True, a `.src.` prefix will be added to the resulting extension
        if it's equal to the destination extension.
        """
        ext = os.path.splitext(self.source_path)[1]
        # do not publish PHP sources
        if prefix and ext == '.html':
            # ext starts with a dot
            return '.src' + ext
        else:
            return ext

    def should_hide_title(self):
        """Return True if this post's title should be hidden. Use in templates to manage posts without titles."""
        return self.title().strip() in ('NO TITLE', '') or self.meta('hidetitle') or \
            self.meta('type').strip() in self.types_to_hide_title

    def should_show_title(self):
        """Return True if this post's title should be displayed. Use in templates to manage posts without titles."""
        return not self.should_hide_title()


def get_metadata_from_file(source_path, post, config, lang, metadata_extractors_by):
    """Extract metadata from the file itself, by parsing contents."""
    try:
        if lang and config:
            source_path = get_translation_candidate(config, source_path, lang)
        elif lang:
            source_path += '.' + lang
        with io.open(source_path, "r", encoding="utf-8-sig") as meta_file:
            source_text = meta_file.read()
    except (UnicodeDecodeError, UnicodeEncodeError):
        msg = 'Error reading {0}: Nikola only supports UTF-8 files'.format(source_path)
        LOGGER.error(msg)
        raise ValueError(msg)
    except Exception:  # The file may not exist, for multilingual sites
        return {}, None

    meta = {}
    used_extractor = None
    for priority in metadata_extractors.MetaPriority:
        found_in_priority = False
        for extractor in metadata_extractors_by['priority'].get(priority, []):
            if not metadata_extractors.check_conditions(post, source_path, extractor.conditions, config, source_text):
                continue
            extractor.check_requirements()
            new_meta = extractor.extract_text(source_text)
            if new_meta:
                found_in_priority = True
                used_extractor = extractor
                # Map metadata from other platforms to names Nikola expects (Issue #2817)
                # Map metadata values (Issue #3025)
                map_metadata(new_meta, extractor.map_from, config)

                meta.update(new_meta)
                break

        if found_in_priority:
            break
    return meta, used_extractor


def get_metadata_from_meta_file(path, post, config, lang, metadata_extractors_by=None):
    """Take a post path, and gets data from a matching .meta file."""
    meta_path = os.path.splitext(path)[0] + '.meta'
    if lang and config:
        meta_path = get_translation_candidate(config, meta_path, lang)
    elif lang:
        meta_path += '.' + lang
    if os.path.isfile(meta_path):
        return get_metadata_from_file(meta_path, post, config, lang, metadata_extractors_by)
    elif lang:
        # Metadata file doesn't exist, but not default language,
        # So, if default language metadata exists, return that.
        # This makes the 2-file format detection more reliable (Issue #525)
        return get_metadata_from_meta_file(meta_path, post, config, None, metadata_extractors_by)
    else:  # No 2-file metadata
        return {}, None


def get_meta(post, lang):
    """Get post meta from compiler or source file."""
    meta = defaultdict(lambda: '')
    used_extractor = None

    config = getattr(post, 'config', None)
    metadata_extractors_by = getattr(post, 'metadata_extractors_by')
    if metadata_extractors_by is None:
        metadata_extractors_by = metadata_extractors.default_metadata_extractors_by()

    # If meta file exists, use it
    metafile_meta, used_extractor = get_metadata_from_meta_file(post.metadata_path, post, config, lang, metadata_extractors_by)

    is_two_file = bool(metafile_meta)

    # Filename-based metadata extractors (priority 1).
    if config.get('FILE_METADATA_REGEXP'):
        extractors = metadata_extractors_by['source'].get(metadata_extractors.MetaSource.filename, [])
        for extractor in extractors:
            if not metadata_extractors.check_conditions(post, post.source_path, extractor.conditions, config, None):
                continue
            meta.update(extractor.extract_filename(post.source_path, lang))

    # Fetch compiler metadata (priority 2, overrides filename-based metadata).
    compiler_meta = {}

    if (getattr(post, 'compiler', None) and post.compiler.supports_metadata and
            metadata_extractors.check_conditions(post, post.source_path, post.compiler.metadata_conditions, config, None)):
        compiler_meta = post.compiler.read_metadata(post, lang=lang)
        used_extractor = post.compiler
        meta.update(compiler_meta)

    # Meta files and inter-file metadata (priority 3, overrides compiler and filename-based metadata).
    if not metafile_meta:
        new_meta, used_extractor = get_metadata_from_file(post.source_path, post, config, lang, metadata_extractors_by)
        meta.update(new_meta)
    else:
        meta.update(metafile_meta)

    if lang is None:
        # Only perform these checks for the default language
        if 'slug' not in meta or not meta['slug']:
            # If no slug is found in the metadata use the filename
            meta['slug'] = slugify(os.path.splitext(
                os.path.basename(post.source_path))[0], post.default_lang)

        if 'title' not in meta or not meta['title']:
            # If no title is found, use the filename without extension
            meta['title'] = os.path.splitext(
                os.path.basename(post.source_path))[0]

    # Set one-file status basing on default language only (Issue #3191)
    if is_two_file or lang is None:
        # Direct access because setter is complicated
        post._is_two_file = is_two_file

    return meta, used_extractor


def hyphenate(dom, _lang):
    """Hyphenate a post."""
    # circular import prevention
    from .nikola import LEGAL_VALUES
    lang = None
    if pyphen is not None:
        lang = LEGAL_VALUES['PYPHEN_LOCALES'].get(_lang, pyphen.language_fallback(_lang))
    else:
        utils.req_missing(['pyphen'], 'hyphenate texts', optional=True)
    hyphenator = None
    if pyphen is not None and lang is not None:
        # If pyphen does exist, we tell the user when configuring the site.
        # If it does not support a language, we ignore it quietly.
        try:
            hyphenator = pyphen.Pyphen(lang=lang)
        except KeyError:
            LOGGER.error("Cannot find hyphenation dictoniaries for {0} (from {1}).".format(lang, _lang))
            LOGGER.error("Pyphen cannot be installed to ~/.local (pip install --user).")
    if hyphenator is not None:
        for tag in ('p', 'li', 'span'):
            for node in dom.xpath("//%s[not(parent::pre)]" % tag):
                skip_node = False
                skippable_nodes = ['kbd', 'pre', 'code', 'samp', 'mark', 'math', 'data', 'ruby', 'svg']
                if node.getchildren():
                    for child in node.getchildren():
                        if child.tag in skippable_nodes or (child.tag == 'span' and 'math'
                                                            in child.get('class', [])):
                            skip_node = True
                elif 'math' in node.get('class', []):
                    skip_node = True
                if not skip_node:
                    insert_hyphens(node, hyphenator)
    return dom


def insert_hyphens(node, hyphenator):
    """Insert hyphens into a node."""
    textattrs = ('text', 'tail')
    if isinstance(node, lxml.etree._Entity):
        # HTML entities have no .text
        textattrs = ('tail',)
    for attr in textattrs:
        text = getattr(node, attr)
        if not text:
            continue

        lines = text.splitlines()
        new_data = "\n".join(
            [
                " ".join([hyphenator.inserted(w, hyphen="\u00AD") for w in line.split(" ")])
                for line in lines
            ]
        )
        # Spaces are trimmed, we have to add them manually back
        if text[0].isspace():
            new_data = ' ' + new_data
        if text[-1].isspace():
            new_data += ' '
        setattr(node, attr, new_data)

    for child in node.iterchildren():
        insert_hyphens(child, hyphenator)



"""Unix support for tzlocal."""
import os
import re

import dateutil.tz

_cache_tz = None


def _try_tz_from_env():
    tzenv = os.environ.get("TZ")
    if tzenv and tzenv[0] == ":":
        tzenv = tzenv[1:]
    try:
        if tzenv:
            dateutil.tz.gettz(tzenv)
            return tzenv
    except Exception:
        pass


def _get_localzone(_root="/"):
    """Try to find the local timezone configuration.

    The parameter _root makes the function look for files like /etc/localtime
    beneath the _root directory. This is primarily used by the tests.
    In normal usage you call the function without parameters.
    """
    tzenv = _try_tz_from_env()
    if tzenv:
        return tzenv

    # Are we under Termux on Android?
    if os.path.exists("/system/bin/getprop"):
        import subprocess

        androidtz = (
            subprocess.check_output(["getprop", "persist.sys.timezone"])
            .strip()
            .decode()
        )
        return androidtz

    # Now look for distribution specific configuration files
    # that contain the timezone name.
    for configfile in ("etc/timezone", "var/db/zoneinfo"):
        tzpath = os.path.join(_root, configfile)
        try:
            with open(tzpath, "rb") as tzfile:
                data = tzfile.read()

                # Issue #3 was that /etc/timezone was a zoneinfo file.
                # That's a misconfiguration, but we need to handle it gracefully:
                if data[:5] == b"TZif2":
                    continue

                etctz = data.strip().decode()
                if not etctz:
                    # Empty file, skip
                    continue
                for etctz in data.decode().splitlines():
                    # Get rid of host definitions and comments:
                    if " " in etctz:
                        etctz, dummy = etctz.split(" ", 1)
                    if "#" in etctz:
                        etctz, dummy = etctz.split("#", 1)
                    if not etctz:
                        continue
                    tz = etctz.replace(" ", "_")
                    return tz

        except IOError:
            # File doesn't exist or is a directory
            continue

    # CentOS has a ZONE setting in /etc/sysconfig/clock,
    # OpenSUSE has a TIMEZONE setting in /etc/sysconfig/clock and
    # Gentoo has a TIMEZONE setting in /etc/conf.d/clock
    # We look through these files for a timezone:

    zone_re = re.compile(r"\s*ZONE\s*=\s*\"")
    timezone_re = re.compile(r"\s*TIMEZONE\s*=\s*\"")
    end_re = re.compile('"')

    for filename in ("etc/sysconfig/clock", "etc/conf.d/clock"):
        tzpath = os.path.join(_root, filename)
        try:
            with open(tzpath, "rt") as tzfile:
                data = tzfile.readlines()

            for line in data:
                # Look for the ZONE= setting.
                match = zone_re.match(line)
                if match is None:
                    # No ZONE= setting. Look for the TIMEZONE= setting.
                    match = timezone_re.match(line)
                if match is not None:
                    # Some setting existed
                    line = line[match.end():]
                    etctz = line[: end_re.search(line).start()]

                    # We found a timezone
                    tz = etctz.replace(" ", "_")
                    return tz

        except IOError:
            # File doesn't exist or is a directory
            continue

    # systemd distributions use symlinks that include the zone name,
    # see manpage of localtime(5) and timedatectl(1)
    tzpath = os.path.join(_root, "etc/localtime")
    if os.path.exists(tzpath) and os.path.islink(tzpath):
        tzpath = os.path.realpath(tzpath)
        start = tzpath.find("/") + 1
        while start != 0:
            tzpath = tzpath[start:]
            try:
                tested_tz = dateutil.tz.gettz(tzpath)
                if tested_tz:
                    return tzpath
            except Exception:
                pass
            start = tzpath.find("/") + 1

    # Nothing found, return UTC
    return None


def get_localzone():
    """Get the computers configured local timezone, if any."""
    global _cache_tz
    if _cache_tz is None:
        _cache_tz = _get_localzone()

    return _cache_tz


def reload_localzone():
    """Reload the cached localzone. You need to call this if the timezone has changed."""
    global _cache_tz
    _cache_tz = _get_localzone()
    return _cache_tz



# -*- encoding:utf-8 -*-
"""Autogenerated file, do not edit. Submit translations on Transifex."""

MESSAGES = {
    "%d min remaining to read": "尚餘 %d 分鐘",
    "(active)": "(啟用)",
    "Also available in:": "其他語言版本：",
    "Archive": "彙整",
    "Atom feed": "",
    "Authors": "作者",
    "Categories": "分類",
    "Comments": "迴響",
    "LANGUAGE": "繁體中文",
    "Languages:": "語言：",
    "More posts about %s": "更多 %s 的文章",
    "Newer posts": "較新的文章",
    "Next post": "下一篇",
    "Next": "下一篇",
    "No posts found.": "沒有找到文章。",
    "Nothing found.": "沒有找到。",
    "Older posts": "較舊的文章",
    "Original site": "原始網站",
    "Posted:": "發佈於：",
    "Posts about %s": "文章分類：%s",
    "Posts by %s": "%s 發佈",
    "Posts for year %s": "%s 年的文章",
    "Posts for {month_day_year}": "{month_day_year}的文章",
    "Posts for {month_year}": "{month_year}的文章",
    "Previous post": "上一篇",
    "Previous": "上一篇",
    "Publication date": "發佈日期",
    "RSS feed": "RSS 訂閱",
    "Read in English": "繁體中文版",
    "Read more": "閱讀更多",
    "Skip to main content": "略過主內容",
    "Source": "原始碼",
    "Subcategories:": "子分類",
    "Tags and Categories": "標籤和分類",
    "Tags": "標籤",
    "Toggle navigation": "切換導航",
    "Uncategorized": "未分類",
    "Up": "向上",
    "Updates": "更新",
    "Write your page here.": "從這裡開始編輯頁面",
    "Write your post here.": "從這裡開始編輯文章",
    "old posts, page %d": "舊文章，第 %d 頁",
    "page %d": "第 %d 頁",
    "updated": "",
}



# -*- coding: utf-8 -*-

# Copyright © 2012-2024 Roberto Alsina and others.

# Permission is hereby granted, free of charge, to any
# person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the
# Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice
# shall be included in all copies or substantial portions of
# the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
# OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""Deploy site."""

import subprocess
import time
from datetime import datetime

import dateutil
from blinker import signal
from dateutil.tz import gettz

from nikola.plugin_categories import Command
from nikola.utils import clean_before_deployment


class CommandDeploy(Command):
    """Deploy site."""

    name = "deploy"

    doc_usage = "[preset [preset...]]"
    doc_purpose = "deploy the site"
    doc_description = "Deploy the site by executing deploy commands from the presets listed on the command line.  If no presets are specified, `default` is executed."

    def _execute(self, command, args):
        """Execute the deploy command."""
        # Get last-deploy from persistent state
        last_deploy = self.site.state.get('last_deploy')
        if last_deploy is not None:
            last_deploy = dateutil.parser.parse(last_deploy)
            clean = False
        else:
            clean = True

        if self.site.config['COMMENT_SYSTEM'] and self.site.config['COMMENT_SYSTEM_ID'] == 'nikolademo':
            self.logger.warning("\nWARNING WARNING WARNING WARNING\n"
                                "You are deploying using the nikolademo Disqus account.\n"
                                "That means you will not be able to moderate the comments in your own site.\n"
                                "And is probably not what you want to do.\n"
                                "Think about it for 5 seconds, I'll wait :-)\n"
                                "(press Ctrl+C to abort)\n")
            time.sleep(5)

        # Remove drafts and future posts if requested
        undeployed_posts = clean_before_deployment(self.site)
        if undeployed_posts:
            self.logger.warning("Deleted {0} posts due to DEPLOY_* settings".format(len(undeployed_posts)))

        if args:
            presets = args
        else:
            presets = ['default']

        # test for preset existence
        for preset in presets:
            try:
                self.site.config['DEPLOY_COMMANDS'][preset]
            except KeyError:
                self.logger.error('No such preset: {0}'.format(preset))
                return 255

        for preset in presets:
            self.logger.info("=> preset '{0}'".format(preset))
            for command in self.site.config['DEPLOY_COMMANDS'][preset]:
                self.logger.info("==> {0}".format(command))
                try:
                    subprocess.check_call(command, shell=True)
                except subprocess.CalledProcessError as e:
                    self.logger.error('Failed deployment -- command {0} '
                                      'returned {1}'.format(e.cmd, e.returncode))
                    return e.returncode

        self.logger.info("Successful deployment")

        new_deploy = datetime.utcnow()
        if last_deploy is None:
            last_deploy = new_deploy
        self._emit_deploy_event(last_deploy, new_deploy, clean, undeployed_posts)

        # Store timestamp of successful deployment
        self.site.state.set('last_deploy', new_deploy.isoformat())
        if clean:
            self.logger.info(
                'Looks like this is the first time you deployed this site. '
                'Let us know you are using Nikola '
                'at <https://users.getnikola.com/add/> if you want!')

    def _emit_deploy_event(self, last_deploy, new_deploy, clean=False, undeployed=None):
        """Emit events for all timeline entries newer than last deploy.

        last_deploy: datetime
            Time stamp of the last successful deployment.

        new_deploy: datetime
            Time stamp of the current deployment.

        clean: bool
            True when it appears like deploy is being run after a clean.

        """
        event = {
            'last_deploy': last_deploy,
            'new_deploy': new_deploy,
            'clean': clean,
            'undeployed': undeployed
        }

        if last_deploy.tzinfo is None:
            last_deploy = last_deploy.replace(tzinfo=gettz('UTC'))

        deployed = [
            entry for entry in self.site.timeline
            if entry.date > last_deploy and entry not in undeployed
        ]

        event['deployed'] = deployed

        if len(deployed) > 0 or len(undeployed) > 0:
            signal('deployed').send(event)



# -*- coding: utf-8 -*-

# Copyright © 2012-2024 Roberto Alsina and others.

# Permission is hereby granted, free of charge, to any
# person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the
# Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice
# shall be included in all copies or substantial portions of
# the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
# OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""Persistent state implementation."""

import json
import os
import shutil
import tempfile
import threading

from . import utils


class Persistor():
    """Persist stuff in a place.

    This is an intentionally dumb implementation. It is *not* meant to be
    fast, or useful for arbitrarily large data. Use lightly.

    Intentionally it has no namespaces, sections, etc. Use as a
    responsible adult.
    """

    def __init__(self, path):
        """Where do you want it persisted."""
        self._path = path
        self._local = threading.local()
        self._local.data = {}

    def _set_site(self, site):
        """Set site and create path directory."""
        self._site = site
        utils.makedirs(os.path.dirname(self._path))

    def get(self, key):
        """Get data stored in key."""
        self._read()
        return self._local.data.get(key)

    def set(self, key, value):
        """Store value in key."""
        self._read()
        self._local.data[key] = value
        self._save()

    def delete(self, key):
        """Delete key and the value it contains."""
        self._read()
        if key in self._local.data:
            self._local.data.pop(key)
        self._save()

    def _read(self):
        if os.path.isfile(self._path):
            with open(self._path) as inf:
                self._local.data = json.load(inf)

    def _save(self):
        dname = os.path.dirname(self._path)
        with tempfile.NamedTemporaryFile(dir=dname, delete=False, mode='w+', encoding='utf-8') as outf:
            tname = outf.name
            json.dump(self._local.data, outf, sort_keys=True, indent=2)
        shutil.move(tname, self._path)



# -*- coding: utf-8 -*-

# Copyright © 2012-2024 Roberto Alsina and others.

# Permission is hereby granted, free of charge, to any
# person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the
# Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice
# shall be included in all copies or substantial portions of
# the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
# OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""Support for Hugo-style shortcodes."""

import sys
import uuid

from .utils import LOGGER


class ParsingError(Exception):
    """Used for forwarding parsing error messages to apply_shortcodes."""

    pass


def _format_position(data, pos):
    """Return position formatted as line/column.

    This is used for prettier error messages.
    """
    line = 0
    col = 0
    llb = ''  # last line break
    for c in data[:pos]:
        if c == '\r' or c == '\n':
            if llb and c != llb:
                llb = ''
            else:
                line += 1
                col = 0
                llb = c
        else:
            col += 1
            llb = ''
    return "line {0}, column {1}".format(line + 1, col + 1)


def _skip_whitespace(data, pos, must_be_nontrivial=False):
    """Return first position after whitespace.

    If must_be_nontrivial is set to True, raises ParsingError
    if no whitespace is found.
    """
    if must_be_nontrivial:
        if pos == len(data) or not data[pos].isspace():
            raise ParsingError("Expecting whitespace at {0}!".format(_format_position(data, pos)))
    while pos < len(data):
        if not data[pos].isspace():
            break
        pos += 1
    return pos


def _skip_nonwhitespace(data, pos):
    """Return first position not before pos which contains a non-whitespace character."""
    for i, x in enumerate(data[pos:]):
        if x.isspace():
            return pos + i
    return len(data)


def _parse_quoted_string(data, start):
    """Parse a quoted string starting at position start in data.

    Returns the position after the string followed by the string itself.
    """
    value = ''
    qc = data[start]
    pos = start + 1
    while pos < len(data):
        char = data[pos]
        if char == '\\':
            if pos + 1 < len(data):
                value += data[pos + 1]
                pos += 2
            else:
                raise ParsingError("Unexpected end of data while escaping ({0})".format(_format_position(data, pos)))
        elif (char == "'" or char == '"') and char == qc:
            return pos + 1, value
        else:
            value += char
            pos += 1
    raise ParsingError("Unexpected end of unquoted string (started at {0})!".format(_format_position(data, start)))


def _parse_unquoted_string(data, start, stop_at_equals):
    """Parse an unquoted string starting at position start in data.

    Returns the position after the string followed by the string itself.
    In case stop_at_equals is set to True, an equal sign will terminate
    the string.
    """
    value = ''
    pos = start
    while pos < len(data):
        char = data[pos]
        if char == '\\':
            if pos + 1 < len(data):
                value += data[pos + 1]
                pos += 2
            else:
                raise ParsingError("Unexpected end of data while escaping ({0})".format(_format_position(data, pos)))
        elif char.isspace():
            break
        elif char == '=' and stop_at_equals:
            break
        elif char == "'" or char == '"':
            raise ParsingError("Unexpected quotation mark in unquoted string ({0})".format(_format_position(data, pos)))
        else:
            value += char
            pos += 1
    return pos, value


def _parse_string(data, start, stop_at_equals=False, must_have_content=False):
    """Parse a string starting at position start in data.

    Returns the position after the string, followed by the string itself, and
    followed by a flog indicating whether the following character is an equals
    sign (only set if stop_at_equals is True).

    If must_have_content is set to True, no empty unquoted strings are accepted.
    """
    if start == len(data):
        raise ParsingError("Expecting string, but found end of input!")
    char = data[start]
    if char == '"' or char == "'":
        end, value = _parse_quoted_string(data, start)
        has_content = True
    else:
        end, value = _parse_unquoted_string(data, start, stop_at_equals)
        has_content = len(value) > 0
    if must_have_content and not has_content:
        raise ParsingError("String starting at {0} must be non-empty!".format(_format_position(data, start)))

    next_is_equals = False
    if stop_at_equals and end + 1 < len(data):
        next_is_equals = (data[end] == '=')
    return end, value, next_is_equals


def _parse_shortcode_args(data, start, shortcode_name, start_pos):
    """When pointed to after a shortcode's name in a shortcode tag, parses the shortcode's arguments until '%}}'.

    Returns the position after '%}}', followed by a tuple (args, kw).

    name and start_pos are only used for formatting error messages.
    """
    args = []
    kwargs = {}

    pos = start
    while True:
        # Skip whitespaces
        try:
            pos = _skip_whitespace(data, pos, must_be_nontrivial=True)
        except ParsingError:
            if not args and not kwargs:
                raise ParsingError("Shortcode '{0}' starting at {1} is not terminated correctly with '%}}}}'!".format(shortcode_name, _format_position(data, start_pos)))
            else:
                raise ParsingError("Syntax error in shortcode '{0}' at {1}: expecting whitespace!".format(shortcode_name, _format_position(data, pos)))
        if pos == len(data):
            break
        # Check for end of shortcode
        if pos + 3 <= len(data) and data[pos:pos + 3] == '%}}':
            return pos + 3, (args, kwargs)
        # Read name
        pos, name, next_is_equals = _parse_string(data, pos, stop_at_equals=True, must_have_content=True)
        if next_is_equals:
            # Read value
            pos, value, _ = _parse_string(data, pos + 1, stop_at_equals=False, must_have_content=False)
            # Store keyword argument
            kwargs[name] = value
        else:
            # Store positional argument
            args.append(name)

    raise ParsingError("Shortcode '{0}' starting at {1} is not terminated correctly with '%}}}}'!".format(shortcode_name, _format_position(data, start_pos)))


def _new_sc_id():
    return str('SHORTCODE{0}REPLACEMENT'.format(str(uuid.uuid4()).replace('-', '')))


def extract_shortcodes(data):
    """
    Return data with replaced shortcodes, shortcodes.

    data is the original data, with the shortcodes replaced by UUIDs.

    a dictionary of shortcodes, where the keys are UUIDs and the values
    are the shortcodes themselves ready to process.
    """
    shortcodes = {}
    splitted = _split_shortcodes(data)

    if not data:  # Empty
        return '', {}

    def extract_data_chunk(data):
        """Take a list of splitted shortcodes and return a string and a tail.

        The string is data, the tail is ready for a new run of this same function.
        """
        text = []
        for i, token in enumerate(data):
            if token[0] == 'SHORTCODE_START':
                name = token[3]
                sc_id = _new_sc_id()
                text.append(sc_id)
                # See if this shortcode closes
                for j in range(i, len(data)):
                    if data[j][0] == 'SHORTCODE_END' and data[j][3] == name:
                        # Extract this chunk
                        shortcodes[sc_id] = ''.join(t[1] for t in data[i:j + 1])
                        return ''.join(text), data[j + 1:]
                # Doesn't close
                shortcodes[sc_id] = token[1]
                return ''.join(text), data[i + 1:]
            elif token[0] == 'TEXT':
                text.append(token[1])
                return ''.join(text), data[1:]
            elif token[0] == 'SHORTCODE_END':  # This is malformed
                raise Exception('Closing unopened shortcode {}'.format(token[3]))

    text = []
    tail = splitted
    while True:
        new_text, tail = extract_data_chunk(tail)
        text.append(new_text)
        if not tail:
            break
    return ''.join(text), shortcodes


def _split_shortcodes(data):
    """Given input data, splits it into a sequence of texts, shortcode starts and shortcode ends.

    Returns a list of tuples of the following forms:

        1. ("TEXT", text)
        2. ("SHORTCODE_START", text, start, name, args)
        3. ("SHORTCODE_END", text, start, name)

    Here, text is the raw text represented by the token; start is the starting position in data
    of the token; name is the name of the shortcode; and args is a tuple (args, kw) as returned
    by _parse_shortcode_args.
    """
    pos = 0
    result = []
    while pos < len(data):
        # Search for shortcode start
        start = data.find('{{%', pos)
        if start < 0:
            result.append(("TEXT", data[pos:]))
            break
        result.append(("TEXT", data[pos:start]))
        # Extract name
        name_start = _skip_whitespace(data, start + 3)
        name_end = _skip_nonwhitespace(data, name_start)
        name = data[name_start:name_end]
        if not name:
            raise ParsingError("Syntax error: '{{{{%' must be followed by shortcode name ({0})!".format(_format_position(data, start)))
        # Finish shortcode
        if name[0] == '/':
            # This is a closing shortcode
            name = name[1:]
            end_start = _skip_whitespace(data, name_end)  # start of '%}}'
            pos = end_start + 3
            # Must be followed by '%}}'
            if pos > len(data) or data[end_start:pos] != '%}}':
                raise ParsingError("Syntax error: '{{{{% /{0}' must be followed by ' %}}}}' ({1})!".format(name, _format_position(data, end_start)))
            result.append(("SHORTCODE_END", data[start:pos], start, name))
        elif name == '%}}':
            raise ParsingError("Syntax error: '{{{{%' must be followed by shortcode name ({0})!".format(_format_position(data, start)))
        else:
            # This is an opening shortcode
            pos, args = _parse_shortcode_args(data, name_end, shortcode_name=name, start_pos=start)
            result.append(("SHORTCODE_START", data[start:pos], start, name, args))
    return result


def apply_shortcodes(data, registry, site=None, filename=None, raise_exceptions=False, lang=None, extra_context=None):
    """Apply Hugo-style shortcodes on data.

    {{% name parameters %}} will end up calling the registered "name" function with the given parameters.
    {{% name parameters %}} something {{% /name %}} will call name with the parameters and
    one extra "data" parameter containing " something ".

    If raise_exceptions is set to True, instead of printing error messages and terminating, errors are
    passed on as exceptions to the caller.

    The site parameter is passed with the same name to the shortcodes so they can access Nikola state.

    >>> print(apply_shortcodes('==> {{% foo bar=baz %}} <==', {'foo': lambda *a, **k: k['bar']}))
    ==> baz <==
    >>> print(apply_shortcodes('==> {{% foo bar=baz %}}some data{{% /foo %}} <==', {'foo': lambda *a, **k: k['bar']+k['data']}))
    ==> bazsome data <==
    """
    if extra_context is None:
        extra_context = {}
    empty_string = ''
    try:
        # Split input data into text, shortcodes and shortcode endings
        sc_data = _split_shortcodes(data)
        # Now process data
        result = []
        dependencies = []
        pos = 0
        while pos < len(sc_data):
            current = sc_data[pos]
            if current[0] == "TEXT":
                result.append(current[1])
                pos += 1
            elif current[0] == "SHORTCODE_END":
                raise ParsingError("Found shortcode ending '{{{{% /{0} %}}}}' which isn't closing a started shortcode ({1})!".format(current[3], _format_position(data, current[2])))
            elif current[0] == "SHORTCODE_START":
                name = current[3]
                # Check if we can find corresponding ending
                found = None
                for p in range(pos + 1, len(sc_data)):
                    if sc_data[p][0] == "SHORTCODE_END" and sc_data[p][3] == name:
                        found = p
                        break
                if found:
                    # Found ending. Extract data argument:
                    data_arg = []
                    for p in range(pos + 1, found):
                        data_arg.append(sc_data[p][1])
                    data_arg = empty_string.join(data_arg)
                    pos = found + 1
                else:
                    # Single shortcode
                    pos += 1
                    data_arg = ''
                args, kw = current[4]
                kw['site'] = site
                kw['data'] = data_arg
                kw['lang'] = lang
                kw.update(extra_context)
                if name in registry:
                    f = registry[name]
                    if getattr(f, 'nikola_shortcode_pass_filename', None):
                        kw['filename'] = filename
                    res = f(*args, **kw)
                    if not isinstance(res, tuple):  # For backards compatibility
                        res = (res, [])
                else:
                    LOGGER.error('Unknown shortcode %s (started at %s)', name, _format_position(data, current[2]))
                    res = ('', [])
                result.append(res[0])
                dependencies += res[1]
        return empty_string.join(result), dependencies
    except ParsingError as e:
        if raise_exceptions:
            # Throw up
            raise
        if filename:
            LOGGER.error("Shortcode error in file {0}: {1}".format(filename, e))
        else:
            LOGGER.error("Shortcode error: {0}".format(e))
        sys.exit(1)



# -*- coding: utf-8 -*-

# Copyright © 2012-2024 Roberto Alsina and others.

# Permission is hereby granted, free of charge, to any
# person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the
# Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice
# shall be included in all copies or substantial portions of
# the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
# OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""Render the taxonomy overviews, classification pages and feeds."""

import os
from collections import defaultdict
from copy import copy
from urllib.parse import urljoin

import blinker
import natsort

from nikola import utils, hierarchy_utils
from nikola.nikola import _enclosure
from nikola.plugin_categories import Task


class RenderTaxonomies(Task):
    """Render taxonomy pages and feeds."""

    name = "render_taxonomies"

    def _generate_classification_overview_kw_context(self, taxonomy, lang):
        """Create context and kw for a classification overview page."""
        context, kw = taxonomy.provide_overview_context_and_uptodate(lang)

        context = copy(context)
        context["kind"] = "{}_index".format(taxonomy.classification_name)
        sorted_links = []
        for other_lang in sorted(self.site.config['TRANSLATIONS'].keys()):
            if other_lang != lang:
                sorted_links.append((other_lang, None, None))
        # Put the current language in front, so that it appears first in links
        # (Issue #3248)
        sorted_links_all = [(lang, None, None)] + sorted_links
        context['has_other_languages'] = True
        context['other_languages'] = sorted_links
        context['all_languages'] = sorted_links_all

        kw = copy(kw)
        kw["messages"] = self.site.MESSAGES
        kw["translations"] = self.site.config['TRANSLATIONS']
        kw["filters"] = self.site.config['FILTERS']
        kw["minimum_post_count"] = taxonomy.minimum_post_count_per_classification_in_overview
        kw["output_folder"] = self.site.config['OUTPUT_FOLDER']
        kw["pretty_urls"] = self.site.config['PRETTY_URLS']
        kw["strip_indexes"] = self.site.config['STRIP_INDEXES']
        kw["index_file"] = self.site.config['INDEX_FILE']

        # Collect all relevant classifications
        if taxonomy.has_hierarchy:
            def acceptor(node):
                return len(self._filter_list(self.site.posts_per_classification[taxonomy.classification_name][lang][node.classification_name], lang)) >= kw["minimum_post_count"]

            clipped_root_list = [hierarchy_utils.clone_treenode(node, parent=None, acceptor=acceptor) for node in self.site.hierarchy_per_classification[taxonomy.classification_name][lang]]
            clipped_root_list = [node for node in clipped_root_list if node]
            clipped_flat_hierarchy = hierarchy_utils.flatten_tree_structure(clipped_root_list)

            classifications = [cat.classification_name for cat in clipped_flat_hierarchy]
        else:
            classifications = natsort.natsorted([tag for tag, posts in self.site.posts_per_classification[taxonomy.classification_name][lang].items()
                                                 if len(self._filter_list(posts, lang)) >= kw["minimum_post_count"]],
                                                alg=natsort.ns.F | natsort.ns.IC)
            taxonomy.sort_classifications(classifications, lang)

        # Set up classifications in context
        context[taxonomy.overview_page_variable_name] = classifications
        context["has_hierarchy"] = taxonomy.has_hierarchy
        if taxonomy.overview_page_items_variable_name:
            items = [(classification,
                      self.site.link(taxonomy.classification_name, classification, lang))
                     for classification in classifications]
            items_with_postcount = [
                (classification,
                 self.site.link(taxonomy.classification_name, classification, lang),
                 len(self._filter_list(self.site.posts_per_classification[taxonomy.classification_name][lang][classification], lang)))
                for classification in classifications
            ]
            context[taxonomy.overview_page_items_variable_name] = items
            context[taxonomy.overview_page_items_variable_name + "_with_postcount"] = items_with_postcount
        if taxonomy.has_hierarchy and taxonomy.overview_page_hierarchy_variable_name:
            hier_items = [
                (node.name, node.classification_name, node.classification_path,
                 self.site.link(taxonomy.classification_name, node.classification_name, lang),
                 node.indent_levels, node.indent_change_before,
                 node.indent_change_after)
                for node in clipped_flat_hierarchy
            ]
            hier_items_with_postcount = [
                (node.name, node.classification_name, node.classification_path,
                 self.site.link(taxonomy.classification_name, node.classification_name, lang),
                 node.indent_levels, node.indent_change_before,
                 node.indent_change_after,
                 len(node.children),
                 len(self._filter_list(self.site.posts_per_classification[taxonomy.classification_name][lang][node.classification_name], lang)))
                for node in clipped_flat_hierarchy
            ]
            context[taxonomy.overview_page_hierarchy_variable_name] = hier_items
            context[taxonomy.overview_page_hierarchy_variable_name + '_with_postcount'] = hier_items_with_postcount
        return context, kw

    def _render_classification_overview(self, classification_name, template, lang, context, kw):
        # Prepare rendering
        context["permalink"] = self.site.link("{}_index".format(classification_name), None, lang)
        if "pagekind" not in context:
            context["pagekind"] = ["list", "tags_page"]
        output_name = os.path.join(self.site.config['OUTPUT_FOLDER'], self.site.path('{}_index'.format(classification_name), None, lang))
        blinker.signal('generate_classification_overview').send({
            'site': self.site,
            'classification_name': classification_name,
            'lang': lang,
            'context': context,
            'kw': kw,
            'output_name': output_name,
        })
        task = self.site.generic_post_list_renderer(
            lang,
            [],
            output_name,
            template,
            kw['filters'],
            context,
        )
        task['uptodate'] = task['uptodate'] + [utils.config_changed(kw, 'nikola.plugins.task.taxonomies:page')]
        task['basename'] = str(self.name)
        yield task

    def _generate_classification_overview(self, taxonomy, lang):
        """Create a global "all your tags/categories" page for a given language."""
        context, kw = self._generate_classification_overview_kw_context(taxonomy, lang)
        for task in self._render_classification_overview(taxonomy.classification_name, taxonomy.template_for_classification_overview, lang, context, kw):
            yield task

    def _generate_tag_and_category_overview(self, tag_taxonomy, category_taxonomy, lang):
        """Create a global "all your tags/categories" page for a given language."""
        # Create individual contexts and kw dicts
        tag_context, tag_kw = self._generate_classification_overview_kw_context(tag_taxonomy, lang)
        cat_context, cat_kw = self._generate_classification_overview_kw_context(category_taxonomy, lang)

        # Combine resp. select dicts
        if tag_context['items'] and cat_context['cat_items']:
            # Combine contexts. We must merge the tag context into the category context
            # so that tag_context['items'] makes it into the result.
            context = cat_context
            context.update(tag_context)
            kw = cat_kw
            kw.update(tag_kw)

            # Update title
            title = self.site.MESSAGES[lang]["Tags and Categories"]
            context['title'] = title
            context['description'] = title
            kw['title'] = title
            kw['description'] = title
        elif cat_context['cat_items']:
            # Use category overview page
            context = cat_context
            kw = cat_kw
        else:
            # Use tag overview page
            context = tag_context
            kw = tag_kw

        # Render result
        for task in self._render_classification_overview('tag', tag_taxonomy.template_for_classification_overview, lang, context, kw):
            yield task

    def _generate_classification_page_as_rss(self, taxonomy, classification, filtered_posts, title, description, kw, lang):
        """Create a RSS feed for a single classification in a given language."""
        kind = taxonomy.classification_name
        # Render RSS
        output_name = os.path.normpath(os.path.join(self.site.config['OUTPUT_FOLDER'], self.site.path(kind + "_rss", classification, lang)))
        feed_url = urljoin(self.site.config['BASE_URL'], self.site.link(kind + "_rss", classification, lang).lstrip('/'))
        deps = []
        deps_uptodate = []
        for post in filtered_posts:
            deps += post.deps(lang)
            deps_uptodate += post.deps_uptodate(lang)
        blog_title = kw["blog_title"](lang)
        task = {
            'basename': str(self.name),
            'name': output_name,
            'file_dep': deps,
            'targets': [output_name],
            'actions': [(utils.generic_rss_renderer,
                        (lang, "{0} ({1})".format(blog_title, title) if blog_title != title else blog_title,
                         kw["site_url"], description, filtered_posts,
                         output_name, kw["feed_teasers"], kw["feed_plain"], kw['feed_length'],
                         feed_url, _enclosure, kw["feed_links_append_query"]))],
            'clean': True,
            'uptodate': [utils.config_changed(kw, 'nikola.plugins.task.taxonomies:rss')] + deps_uptodate,
            'task_dep': ['render_posts'],
        }
        return utils.apply_filters(task, kw['filters'])

    def _generate_classification_page_as_index(self, taxonomy, classification, filtered_posts, context, kw, lang):
        """Render an index page collection using only this classification's posts."""
        kind = taxonomy.classification_name

        def page_link(i, displayed_i, num_pages, force_addition, extension=None):
            return self.site.link(kind, classification, lang, alternative_path=force_addition, page=i)

        def page_path(i, displayed_i, num_pages, force_addition, extension=None):
            return self.site.path(kind, classification, lang, alternative_path=force_addition, page=i)

        context = copy(context)
        context["kind"] = kind
        if "pagekind" not in context:
            context["pagekind"] = ["index", "tag_page"]
        template_name = taxonomy.template_for_single_list

        yield self.site.generic_index_renderer(lang, filtered_posts, context['title'], template_name, context, kw, str(self.name), page_link, page_path)

    def _generate_classification_page_as_atom(self, taxonomy, classification, filtered_posts, context, kw, lang):
        """Generate atom feeds for classification lists."""
        kind = taxonomy.classification_name

        context = copy(context)
        context["kind"] = kind

        yield self.site.generic_atom_renderer(lang, filtered_posts, context, kw, str(self.name), classification, kind)

    def _generate_classification_page_as_list(self, taxonomy, classification, filtered_posts, context, kw, lang):
        """Render a single flat link list with this classification's posts."""
        kind = taxonomy.classification_name
        template_name = taxonomy.template_for_single_list
        output_name = os.path.join(self.site.config['OUTPUT_FOLDER'], self.site.path(kind, classification, lang))
        context["lang"] = lang
        # list.tmpl expects a different format than list_post.tmpl (Issue #2701)
        if template_name == 'list.tmpl':
            context["items"] = [(post.title(lang), post.permalink(lang), None) for post in filtered_posts]
        else:
            context["posts"] = filtered_posts
        if "pagekind" not in context:
            context["pagekind"] = ["list", "tag_page"]
        task = self.site.generic_post_list_renderer(lang, filtered_posts, output_name, template_name, kw['filters'], context)
        task['uptodate'] = task['uptodate'] + [utils.config_changed(kw, 'nikola.plugins.task.taxonomies:list')]
        task['basename'] = str(self.name)
        yield task

    def _filter_list(self, post_list, lang):
        """Return only the posts which should be shown for this language."""
        if self.site.config["SHOW_UNTRANSLATED_POSTS"]:
            return post_list
        else:
            return [x for x in post_list if x.is_translation_available(lang)]

    def _generate_subclassification_page(self, taxonomy, node, context, kw, lang):
        """Render a list of subclassifications."""
        def get_subnode_data(subnode):
            return [
                taxonomy.get_classification_friendly_name(subnode.classification_name, lang, only_last_component=True),
                self.site.link(taxonomy.classification_name, subnode.classification_name, lang),
                len(self._filter_list(self.site.posts_per_classification[taxonomy.classification_name][lang][subnode.classification_name], lang))
            ]

        items = [get_subnode_data(subnode) for subnode in node.children]
        context = copy(context)
        context["lang"] = lang
        context["permalink"] = self.site.link(taxonomy.classification_name, node.classification_name, lang)
        if "pagekind" not in context:
            context["pagekind"] = ["list", "archive_page"]
        context["items"] = items
        task = self.site.generic_post_list_renderer(
            lang,
            [],
            os.path.join(kw['output_folder'], self.site.path(taxonomy.classification_name, node.classification_name, lang)),
            taxonomy.subcategories_list_template,
            kw['filters'],
            context,
        )
        task_cfg = {1: kw, 2: items}
        task['uptodate'] = task['uptodate'] + [utils.config_changed(task_cfg, 'nikola.plugins.task.taxonomy')]
        task['basename'] = self.name
        return task

    def _generate_classification_page(self, taxonomy, classification, filtered_posts, generate_list, generate_rss, generate_atom, lang, post_lists_per_lang, classification_set_per_lang=None):
        """Render index or post list and associated feeds per classification."""
        # Should we create this list?
        if not any((generate_list, generate_rss, generate_atom)):
            return
        # Get data
        node = None
        if taxonomy.has_hierarchy:
            node = self.site.hierarchy_lookup_per_classification[taxonomy.classification_name][lang].get(classification)
        context, kw = taxonomy.provide_context_and_uptodate(classification, lang, node)
        kw = copy(kw)
        kw["messages"] = self.site.MESSAGES
        kw["translations"] = self.site.config['TRANSLATIONS']
        kw["filters"] = self.site.config['FILTERS']
        kw["site_url"] = self.site.config['SITE_URL']
        kw["blog_title"] = self.site.config['BLOG_TITLE']
        kw["generate_rss"] = self.site.config['GENERATE_RSS']
        kw["generate_atom"] = self.site.config['GENERATE_ATOM']
        kw["feed_teasers"] = self.site.config["FEED_TEASERS"]
        kw["feed_plain"] = self.site.config["FEED_PLAIN"]
        kw["feed_links_append_query"] = self.site.config["FEED_LINKS_APPEND_QUERY"]
        kw["feed_length"] = self.site.config['FEED_LENGTH']
        kw["output_folder"] = self.site.config['OUTPUT_FOLDER']
        kw["pretty_urls"] = self.site.config['PRETTY_URLS']
        kw["strip_indexes"] = self.site.config['STRIP_INDEXES']
        kw["index_file"] = self.site.config['INDEX_FILE']
        context = copy(context)
        context["permalink"] = self.site.link(taxonomy.classification_name, classification, lang)
        context["kind"] = taxonomy.classification_name
        # Get links to other language versions of this classification
        if classification_set_per_lang is not None:
            other_lang_links = taxonomy.get_other_language_variants(classification, lang, classification_set_per_lang)
            # Collect by language
            links_per_lang = defaultdict(list)
            for other_lang, link in other_lang_links:
                # Make sure we ignore the current language (in case the
                # plugin accidentally returns links for it as well)
                if other_lang != lang:
                    links_per_lang[other_lang].append(link)
            # Sort first by language, then by classification
            sorted_links = []
            sorted_links_all = []
            for other_lang in sorted(list(links_per_lang.keys()) + [lang]):
                if other_lang == lang:
                    sorted_links_all.append((lang, classification, taxonomy.get_classification_friendly_name(classification, lang)))
                else:
                    links = hierarchy_utils.sort_classifications(taxonomy, links_per_lang[other_lang], other_lang)
                    links = [(other_lang, other_classification,
                              taxonomy.get_classification_friendly_name(other_classification, other_lang))
                             for other_classification in links if post_lists_per_lang[other_lang].get(other_classification, ('', False, False))[1]]
                    sorted_links.extend(links)
                    sorted_links_all.extend(links)
            # Store result in context and kw
            context['has_other_languages'] = True
            context['other_languages'] = sorted_links
            context['all_languages'] = sorted_links_all
            kw['other_languages'] = sorted_links
            kw['all_languages'] = sorted_links_all
        else:
            context['has_other_languages'] = False
        # Allow other plugins to modify the result
        blinker.signal('generate_classification_page').send({
            'site': self.site,
            'taxonomy': taxonomy,
            'classification': classification,
            'lang': lang,
            'posts': filtered_posts,
            'context': context,
            'kw': kw,
        })
        # Decide what to do
        if taxonomy.has_hierarchy and taxonomy.show_list_as_subcategories_list:
            # Determine whether there are subcategories
            node = self.site.hierarchy_lookup_per_classification[taxonomy.classification_name][lang][classification]
            # Are there subclassifications?
            if len(node.children) > 0:
                # Yes: create list with subclassifications instead of list of posts
                if generate_list:
                    yield self._generate_subclassification_page(taxonomy, node, context, kw, lang)
                return
        # Generate RSS feed
        if generate_rss and kw["generate_rss"] and not taxonomy.always_disable_rss:
            yield self._generate_classification_page_as_rss(taxonomy, classification, filtered_posts, context['title'], context.get("description"), kw, lang)

        # Generate Atom feed
        if generate_atom and kw["generate_atom"] and not taxonomy.always_disable_atom:
            yield self._generate_classification_page_as_atom(taxonomy, classification, filtered_posts, context, kw, lang)

        # Render HTML
        if generate_list and taxonomy.show_list_as_index:
            yield self._generate_classification_page_as_index(taxonomy, classification, filtered_posts, context, kw, lang)
        elif generate_list:
            yield self._generate_classification_page_as_list(taxonomy, classification, filtered_posts, context, kw, lang)

    def gen_tasks(self):
        """Render the tag pages and feeds."""
        self.site.scan_posts()
        yield self.group_task()

        # Cache classification sets per language for taxonomies where
        # add_other_languages_variable is True.
        classification_set_per_lang = {}
        for taxonomy in self.site.taxonomy_plugins.values():
            if taxonomy.add_other_languages_variable:
                lookup = self.site.posts_per_classification[taxonomy.classification_name]
                cspl = {lang: set(lookup[lang].keys()) for lang in lookup}
                classification_set_per_lang[taxonomy.classification_name] = cspl

        # Collect post lists for classification pages and determine whether
        # they should be generated.
        post_lists_per_lang = {}
        for taxonomy in self.site.taxonomy_plugins.values():
            plpl = {}
            for lang in self.site.config["TRANSLATIONS"]:
                result = {}
                for classification, posts in self.site.posts_per_classification[taxonomy.classification_name][lang].items():
                    # Filter list
                    filtered_posts = self._filter_list(posts, lang)
                    if len(filtered_posts) == 0 and taxonomy.omit_empty_classifications:
                        generate_list = generate_rss = generate_atom = False
                    else:
                        # Should we create this list?
                        generate_list = taxonomy.should_generate_classification_page(classification, filtered_posts, lang)
                        generate_rss = taxonomy.should_generate_rss_for_classification_page(classification, filtered_posts, lang)
                        generate_atom = taxonomy.should_generate_atom_for_classification_page(classification, filtered_posts, lang)
                    result[classification] = (filtered_posts, generate_list, generate_rss, generate_atom)
                plpl[lang] = result
            post_lists_per_lang[taxonomy.classification_name] = plpl

        # Now generate pages
        for lang in self.site.config["TRANSLATIONS"]:
            # To support that tag and category classifications share the same overview,
            # we explicitly detect this case:
            ignore_plugins_for_overview = set()
            if 'tag' in self.site.taxonomy_plugins and 'category' in self.site.taxonomy_plugins and self.site.link("tag_index", None, lang) == self.site.link("category_index", None, lang):
                # Block both plugins from creating overviews
                ignore_plugins_for_overview.add(self.site.taxonomy_plugins['tag'])
                ignore_plugins_for_overview.add(self.site.taxonomy_plugins['category'])
            for taxonomy in self.site.taxonomy_plugins.values():
                if not taxonomy.is_enabled(lang):
                    continue
                # Generate list of classifications (i.e. classification overview)
                if taxonomy not in ignore_plugins_for_overview:
                    if taxonomy.template_for_classification_overview is not None:
                        for task in self._generate_classification_overview(taxonomy, lang):
                            yield task

                # Process classifications
                for classification, (filtered_posts, generate_list, generate_rss, generate_atom) in post_lists_per_lang[taxonomy.classification_name][lang].items():
                    for task in self._generate_classification_page(taxonomy, classification, filtered_posts,
                                                                   generate_list, generate_rss, generate_atom, lang,
                                                                   post_lists_per_lang[taxonomy.classification_name],
                                                                   classification_set_per_lang.get(taxonomy.classification_name)):
                        yield task
            # In case we are ignoring plugins for overview, we must have a collision for
            # tags and categories. Handle this special case with extra code.
            if ignore_plugins_for_overview:
                for task in self._generate_tag_and_category_overview(self.site.taxonomy_plugins['tag'], self.site.taxonomy_plugins['category'], lang):
                    yield task



# -*- encoding:utf-8 -*-
"""Autogenerated file, do not edit. Submit translations on Transifex."""

MESSAGES = {
    "%d min remaining to read": "%d perc van hátra olvasni",
    "(active)": "(aktív)",
    "Also available in:": "Olvasható még:",
    "Archive": "Archív",
    "Atom feed": "Atom",
    "Authors": "Szerzők",
    "Categories": "Kategóriák",
    "Comments": "Hozzászólások",
    "LANGUAGE": "Magyarul",
    "Languages:": "Nyelvek:",
    "More posts about %s": "Több bejegyzés erről: %s",
    "Newer posts": "Újabb bejegyzések",
    "Next post": "A következő bejegyzés",
    "Next": "Következő",
    "No posts found.": "Nincs ilyen bejegyzés.",
    "Nothing found.": "Nincs találat.",
    "Older posts": "Régebbi bejegyzések",
    "Original site": "Eredeti honlap",
    "Posted:": "Megjelent:",
    "Posts about %s": "Bejegyzések erről: %s",
    "Posts by %s": "Bejegyzések %s által",
    "Posts for year %s": "%s. bejegyzések",
    "Posts for {month_day_year}": "{month_day_year} bejegyzések",
    "Posts for {month_year}": "{month_year} bejegyzések",
    "Previous post": "Az előző bejegyzés ",
    "Previous": "Előző",
    "Publication date": "A megjelenés dátuma",
    "RSS feed": "RSS",
    "Read in English": "Olvass magyarul",
    "Read more": "Olvass tovább",
    "Skip to main content": "Ugrás a főtartalomra",
    "Source": "Forrás",
    "Subcategories:": "Alkategóriák:",
    "Tags and Categories": "Címkék és kategóriák",
    "Tags": "Címkék",
    "Toggle navigation": "",
    "Uncategorized": "Nincs kategorizálva",
    "Up": "Fel",
    "Updates": "Frissítések",
    "Write your page here.": "Ide írd az oldalad.",
    "Write your post here.": "Ide írd a bejegyzésed.",
    "old posts, page %d": "régi bejegyzések, %d. oldal",
    "page %d": "%d. oldal",
    "updated": "",
}



# -*- coding: utf-8 -*-

# Copyright © 2012-2024 Roberto Alsina and others.

# Permission is hereby granted, free of charge, to any
# person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the
# Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice
# shall be included in all copies or substantial portions of
# the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
# OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""Hierarchy utility functions."""

import natsort

__all__ = ('TreeNode', 'clone_treenode', 'flatten_tree_structure',
           'sort_classifications', 'join_hierarchical_category_path',
           'parse_escaped_hierarchical_category_name',)


class TreeNode(object):
    """A tree node."""

    indent_levels = None  # use for formatting comments as tree
    indent_change_before = 0  # use for formatting comments as tree
    indent_change_after = 0  # use for formatting comments as tree

    # The indent levels and changes allow to render a tree structure
    # without keeping track of all that information during rendering.
    #
    # The indent_change_before is the different between the current
    # comment's level and the previous comment's level; if the number
    # is positive, the current level is indented further in, and if it
    # is negative, it is indented further out. Positive values can be
    # used to open HTML tags for each opened level.
    #
    # The indent_change_after is the difference between the next
    # comment's level and the current comment's level. Negative values
    # can be used to close HTML tags for each closed level.
    #
    # The indent_levels list contains one entry (index, count) per
    # level, informing about the index of the current comment on that
    # level and the count of comments on that level (before a comment
    # of a higher level comes). This information can be used to render
    # tree indicators, for example to generate a tree such as:
    #
    # +--- [(0,3)]
    # +-+- [(1,3)]
    # | +--- [(1,3), (0,2)]
    # | +-+- [(1,3), (1,2)]
    # |   +--- [(1,3), (1,2), (0, 1)]
    # +-+- [(2,3)]
    #   +- [(2,3), (0,1)]
    #
    # (The lists used as labels represent the content of the
    # indent_levels property for that node.)

    def __init__(self, name, parent=None):
        """Initialize node."""
        self.name = name
        self.parent = parent
        self.children = []

    def get_path(self):
        """Get path."""
        path = []
        curr = self
        while curr is not None:
            path.append(curr)
            curr = curr.parent
        return reversed(path)

    def get_children(self):
        """Get children of a node."""
        return self.children

    def __str__(self):
        """Stringify node (return name)."""
        return self.name

    def _repr_partial(self):
        """Return partial representation."""
        if self.parent:
            return "{0}/{1!r}".format(self.parent._repr_partial(), self.name)
        else:
            return repr(self.name)

    def __repr__(self):
        """Return programmer-friendly node representation."""
        return "<TreeNode {0}>".format(self._repr_partial())


def clone_treenode(treenode, parent=None, acceptor=lambda x: True):
    """Clone a TreeNode.

    Children are only cloned if `acceptor` returns `True` when
    applied on them.

    Returns the cloned node if it has children or if `acceptor`
    applied to it returns `True`. In case neither applies, `None`
    is returned.
    """
    # Copy standard TreeNode stuff
    node_clone = TreeNode(treenode.name, parent)
    node_clone.children = [clone_treenode(node, parent=node_clone, acceptor=acceptor) for node in treenode.children]
    node_clone.children = [node for node in node_clone.children if node]
    node_clone.indent_levels = treenode.indent_levels
    node_clone.indent_change_before = treenode.indent_change_before
    node_clone.indent_change_after = treenode.indent_change_after
    if hasattr(treenode, 'classification_path'):
        # Copy stuff added by taxonomies_classifier plugin
        node_clone.classification_path = treenode.classification_path
        node_clone.classification_name = treenode.classification_name

    # Accept this node if there are no children (left) and acceptor fails
    if not node_clone.children and not acceptor(treenode):
        return None
    return node_clone


def flatten_tree_structure(root_list):
    """Flatten a tree."""
    elements = []

    def generate(input_list, indent_levels_so_far):
        """Generate flat list of nodes."""
        for index, element in enumerate(input_list):
            # add to destination
            elements.append(element)
            # compute and set indent levels
            indent_levels = indent_levels_so_far + [(index, len(input_list))]
            element.indent_levels = indent_levels
            # add children
            children = element.get_children()
            element.children_count = len(children)
            generate(children, indent_levels)

    generate(root_list, [])
    # Add indent change counters
    level = 0
    last_element = None
    for element in elements:
        new_level = len(element.indent_levels)
        # Compute level change before this element
        change = new_level - level
        if last_element is not None:
            last_element.indent_change_after = change
        element.indent_change_before = change
        # Update variables
        level = new_level
        last_element = element
    # Set level change after last element
    if last_element is not None:
        last_element.indent_change_after = -level
    return elements


def parse_escaped_hierarchical_category_name(category_name):
    """Parse a category name."""
    result = []
    current = None
    index = 0
    next_backslash = category_name.find('\\', index)
    next_slash = category_name.find('/', index)
    while index < len(category_name):
        if next_backslash == -1 and next_slash == -1:
            current = (current if current else "") + category_name[index:]
            index = len(category_name)
        elif next_slash >= 0 and (next_backslash == -1 or next_backslash > next_slash):
            result.append((current if current else "") + category_name[index:next_slash])
            current = ''
            index = next_slash + 1
            next_slash = category_name.find('/', index)
        else:
            if len(category_name) == next_backslash + 1:
                raise Exception("Unexpected '\\' in '{0}' at last position!".format(category_name))
            esc_ch = category_name[next_backslash + 1]
            if esc_ch not in {'/', '\\'}:
                raise Exception("Unknown escape sequence '\\{0}' in '{1}'!".format(esc_ch, category_name))
            current = (current if current else "") + category_name[index:next_backslash] + esc_ch
            index = next_backslash + 2
            next_backslash = category_name.find('\\', index)
            if esc_ch == '/':
                next_slash = category_name.find('/', index)
    if current is not None:
        result.append(current)
    return result


def join_hierarchical_category_path(category_path):
    """Join a category path."""
    def escape(s):
        """Espace one part of category path."""
        return s.replace('\\', '\\\\').replace('/', '\\/')

    return '/'.join([escape(p) for p in category_path])


def sort_classifications(taxonomy, classifications, lang):
    """Sort the given list of classifications of the given taxonomy and language.

    ``taxonomy`` must be a ``Taxonomy`` plugin.
    ``classifications`` must be an iterable collection of
    classification strings for that taxonomy.
    ``lang`` is the language the classifications are for.

    The result will be returned as a sorted list. Sorting will
    happen according to the way the complete classification
    hierarchy for the taxonomy is sorted.
    """
    if taxonomy.has_hierarchy:
        # To sort a hierarchy of classifications correctly, we first
        # build a tree out of them (and mark for each node whether it
        # appears in the list), then sort the tree node-wise, and finally
        # collapse the tree into a list of recombined classifications.

        # Step 1: build hierarchy. Here, each node consists of a boolean
        # flag (node appears in list) and a dictionary mapping path elements
        # to nodes.
        root = [False, {}]
        for classification in classifications:
            node = root
            for elt in taxonomy.extract_hierarchy(classification):
                if elt not in node[1]:
                    node[1][elt] = [False, {}]
                node = node[1][elt]
            node[0] = True
        # Step 2: sort hierarchy. The result for a node is a pair
        # (flag, subnodes), where subnodes is a list of pairs (name, subnode).

        def sort_node(node, level=0):
            """Return sorted node, with children as `(name, node)` list instead of a dictionary."""
            keys = natsort.natsorted(node[1].keys(), alg=natsort.ns.F | natsort.ns.IC)
            taxonomy.sort_classifications(keys, lang, level)
            subnodes = []
            for key in keys:
                subnodes.append((key, sort_node(node[1][key], level + 1)))
            return (node[0], subnodes)

        root = sort_node(root)
        # Step 3: collapse the tree structure into a linear sorted list,
        # with a node coming before its children.

        def append_node(classifications, node, path=()):
            """Append the node and then its children to the classifications list."""
            if node[0]:
                classifications.append(taxonomy.recombine_classification_from_hierarchy(path))
            for key, subnode in node[1]:
                append_node(classifications, subnode, path + (key, ))

        classifications = []
        append_node(classifications, root)
        return classifications
    else:
        # Sorting a flat hierarchy is simpler. We pre-sort with
        # natsorted and call taxonomy.sort_classifications.
        classifications = natsort.natsorted(classifications, alg=natsort.ns.F | natsort.ns.IC)
        taxonomy.sort_classifications(classifications, lang)
        return classifications
